/*
 * Data Loop.  bsy, 94-09-18.
 *
 * Basic tool:  takes two processes as argument, and run them with
 * standard input/outputs hooked together, forming a data flow loop.
 * Caveat emptor:  stdio buffering and pipe capacity limits can easily
 * cause deadlocks.
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#define	DELIMITER "---"

char	*me;

usage()
{
	fprintf(stderr,"Usage:  %s [-d delim_string] process1 [delim_string] process2\n",me);
	fprintf(stderr,"        default delim_string is \"%s\"\n",DELIMITER);
}

main(int ac,char **av)
{
	extern int	getopt(int,char *const*,const char *);
	extern int	optind;
	extern char	*optarg, *strrchr(const char *,char);

	char		*delim = DELIMITER;
	int		opt, ix;
	int		proc1pid, proc2pid, exitstat;
	char		**proc1 = 0, **proc2 = 0;
	int		topipe[2], frompipe[2];

	if (me = strrchr(av[0],'/')) ++me; else me = av[0];
	while ((opt = getopt(ac,av,"d:")) != EOF) switch (opt) {
	case 'd':	delim = optarg;	break;
	default:	usage(); exit(1);
	}
	proc1 = av + optind;
	for (ix = optind; ix < ac; ++ix) {
		if (!strcmp(av[ix],delim)) {
			av[ix] = 0;
			proc2 = av + ix + 1; break;
		}
	}
	if (!*proc1) {
		fprintf(stderr,"%s:  missing process specifications\n",me);
		usage(); exit(1);
	}
	if (!proc2 || !*proc2) {
		fprintf(stderr,"%s:  missing 2nd process specification\n",me);
		usage(); exit(1);
	}
	if (pipe(topipe) == -1 || pipe(frompipe) == -1) {
		perror(me);
		exit(1);
	}
	switch (proc2pid = fork()) {
	case -1:
		perror(me); exit(1);
	case 0:
		(void) dup2(topipe[0],0);
		(void) close(topipe[0]);
		(void) close(topipe[1]);
		(void) dup2(frompipe[1],1);
		(void) close(frompipe[0]);
		(void) close(frompipe[1]);
		(void) execvp(*proc2,proc2);
		perror(me);
		fprintf(stderr,"%s: exec of %s failed\n",me,*proc2);
		_exit(1);
		break;
	}
	switch (proc1pid = fork()) {
	case -1:
		perror(me); (void) kill(proc2pid,SIGKILL);
		exit(1);
	case 0:
		(void) dup2(topipe[1],1);
		(void) close(topipe[0]);
		(void) close(topipe[1]);
		(void) dup2(frompipe[0],0);
		(void) close(frompipe[0]);
		(void) close(frompipe[1]);
		(void) execvp(*proc1,proc1);
		perror(me);
		fprintf(stderr,"%s: exec of %s failed\n",me,*proc1);
		/* no need to kill other process, EOF on stdin should do it */
		_exit(1);
		break;
	}

	(void) close(topipe[0]);
	(void) close(topipe[1]);

	(void) close(frompipe[0]);
	(void) close(frompipe[1]);
	exitstat = wait(0);
	exitstat |= wait(0);
	exit(exitstat);
}

