#ifndef	lint
static char *copyright = "\n\
\n\
Copyright (c) 1994 by Bennet Yee.\n\
\n\
This material was developed by the author.  Permission to copy this\n\
software, to redistribute it, and to use it for any purpose is hereby\n\
granted, subject to the following five restrictions.\n\
\n\
1. Any copy made of this software must include this copyright notice in\n\
full.\n\
\n\
2. All materials containing this software or derivatives thereof must\n\
provide this software in source form or provide a means of obtaining\n\
this software in source form; no fees may be charged for this software\n\
except those for covering media costs and handling.\n\
\n\
3. All materials developed as a consequence of the use of this software\n\
shall duly acknowledge such use, in accordance with the usual standards\n\
of acknowledging credit in academic research.\n\
\n\
4. The author makes no warranty or representation that the operation\n\
of this software will be error-free, and the author is under no\n\
obligation to provide any services, by way of maintenance, update, or\n\
otherwise.\n\
\n\
5. In conjunction with products arising from the use of this material,\n\
there shall be no use of the names of the author, of Carnegie-Mellon\n\
University, nor of any adaptation thereof in any advertising,\n\
promotional, or sales literature without prior written consent from the\n\
author and Carnegie-Mellon University in each case.\n\
\n\
If you have questions, you can contact the author at bsy@cs.cmu.edu\n\
\n\
Users of this software is requested to make their best efforts (a) to\n\
return to the author any improvements or extensions that they make, so\n\
that these may be included in future releases; (b) to document clearly\n\
any improvements or extensions made to this software; and (c) to inform\n\
the author of noteworthy uses of this software.  Any improvements or\n\
extensions to this software may be placed under different copyright\n\
restrictions by the author of those improvements or extensions.  That\n\
those improvements or extensions be placed under the same copyright as\n\
this software is encouraged.\n\
";
#endif
/* lint */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <netdb.h>
#include <signal.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/wait.h>
#include <unistd.h>

#define	BUFSIZE		16383
#define	SRVBUF		1024
#define	PROTOBUF	1024

#define MAXRETRY	10
#define	HOSTNAMEBUF	256

#define	USE_AF_INET	1
#define DEBUG		0

int	debug = 0, quiet = 0, broadcast = 0;
int	add_cr = 1, strip_cr = 1, exit_on_eof = 0;

extern int	errno;
extern int	getopt(), optind;
extern char	*optarg;

/*
 * Different network names are allowed, but since we use sockaddr_in here
 * things probably won't work -- other address families / protocol families
 * are not really supported; this is mostly for future expansion.
 */
char *netname = "cmu-net", *machinename = (char *) 0, hostnamebuf[HOSTNAMEBUF];
char *servicename = (char *) 0, *protoname = (char *) 0;

main(ac,av)
int ac;
char **av;
{
	int			s, socktype, child;
	char			*cp, *index();
	struct servent		*service;
	struct netent		*network;
	struct hostent		*host;
	struct sockaddr_in	saddrin;
	struct protoent		*proto;
	int			opt;
	int			addrtype;

	while ((opt = getopt(ac,av,"bqdeQRn:h:s:p:")) != EOF) switch (opt) {
	case 'b':	broadcast = 1;		break;
	case 'q':	quiet = 1;		break;
	case 'd':	debug = 1;		break;
	case 'e':	exit_on_eof = 1;	break;
	case 'Q':	add_cr = 0;		break;	/* quote input */
	case 'R':	strip_cr = 0;		break;	/* leave returns */
	case 'n':	netname = optarg;	break;
	case 'h':	machinename = optarg;	break;
	case 's':	servicename = optarg;	break;
	case 'p':	protoname = optarg;	break;
	default:
		fprintf(stderr,"server:  unknown switch %c\n",opt);
		usage();
		exit(1);
	}
	if (optind != ac) {
		fprintf(stderr,"server: extraneous arguments %s ...\n",av[optind]);
		usage();
		exit(1);
	}

	if (!machinename) {
		if (gethostname(hostnamebuf,HOSTNAMEBUF) == -1) {
			perror("server:gethostname");
			fprintf(stderr,"server:  where am I?!?\n");
			exit(1);
		}
		machinename = hostnamebuf;
	}
	if (!servicename) {
		fprintf(stderr,"server: must name to which server to connect\n");
		usage(); exit(1);
	}
	if (!protoname) {
		fprintf(stderr,"server: must specify protocol to use\n");
		usage(); exit(1);
	}
	if (isdigit(*servicename)) {
		static struct servent srv;
		service = &srv;
		srv.s_proto = protoname;
		srv.s_port = atoi(servicename);
		srv.s_port = htons(srv.s_port);
	} else {
		if (!(service = getservbyname(servicename,protoname))) {
			fprintf(stderr,"server:  can not find %s service with %s protocol\n",servicename,protoname);
			exit(1);
		}
	}
	if (!(network = getnetbyname(netname))) {
#if	USE_AF_INET
		addrtype = AF_INET;
#else
		fprintf(stderr,"server:  can not get network info for %s\n",netname);
		exit(1);
#endif
	} else {
		addrtype = network->n_addrtype;
	}

	if (!(proto = getprotobyname(service->s_proto))) {
		fprintf(stderr,"server:  can not obtain protocol info for %s\n",service->s_proto);
		exit(1);
	}

	if (!strcmp(service->s_proto,"udp")) {
		socktype = SOCK_DGRAM;
	} else {
		socktype = SOCK_STREAM;
	}

	if (debug) {
		fprintf(stderr,"obtaining a socket\n");
		fprintf(stderr,"protocol number is %d\n",proto->p_proto);
	}

	if ((s = socket(addrtype,socktype,proto->p_proto)) == -1) {
		perror("server:socket");
		fprintf(stderr,"errno = %d\n",errno);
		exit(1);
	}
	if (broadcast) {
		int	on = 1;
		if (setsockopt(s,SOL_SOCKET,SO_BROADCAST,(const char *) &on,sizeof on) == -1) {
			perror("server:setsockopt");
			exit(1);
		}
	}

	memset((char *) &saddrin,0,sizeof saddrin);
	if (!isdigit(*machinename)) {
		if (!(host = gethostbyname(machinename))) {
			fprintf(stderr,"server:  can not get host info for %s\n",machinename);
			exit(1);
		}
		memcpy((char *) &saddrin.sin_addr,host->h_addr,host->h_length);
	} else {
		saddrin.sin_addr.s_addr = inet_addr(machinename);
	}
	saddrin.sin_family = addrtype;
	saddrin.sin_port = service->s_port;

	if (debug) {
		fprintf(stderr,"obtaining a connection with socket %x\n",s);
		fprintf(stderr,"connecting to address family %d, address %X, and port %d\n",saddrin.sin_family,saddrin.sin_addr,ntohs(saddrin.sin_port));
	}

	if (connect(s,(struct sockaddr *) &saddrin,sizeof saddrin) == -1) {
		perror("server:connect");
		fprintf(stderr,"errno = %d\n",errno);
		exit(1);
	}

	if (!quiet) fprintf(stderr,"CONNECTED\n");

	child = doit(s,socktype);
	(void) close(s);
	if (debug) {
		fprintf(stderr,"eof encountered\n");
		fprintf(stderr,"waiting for child death\n");
	}
#if	defined(_POSIX_VERSION)
	while (wait((int *) 0) != child)
		;
#else
	while (wait((union wait *) 0) != child)
		;
#endif
	if (!quiet) fprintf(stderr,"CONNECTION CLOSED\n");
	exit(0);
}

usage()
{
	fprintf(stderr,"server:  usage:  server [-bdeqQR] [-s service] [-p protocol] [-n network] [-h host]\n");
}

char buf[BUFSIZE], cr_buf[2 * BUFSIZE];

doit(s,stype)
int s,stype;
{
	int l, m, ckin(), retry, child;
	char *bufptr;

	switch (child = fork()) {
	case -1:
		perror("server:fork");
		fprintf(stderr,"server:  can not fork read process\n");
		exit(1);
		break;
	case 0:
		ckin(s,stype);
		_exit(0);
	default:
		while ((l = read(0,buf,BUFSIZE)) > 0) {
			if (debug) {
				fprintf(stderr,"sending \"");
				fflush(stderr);
				write(2,buf,l);
				fprintf(stderr,"\"\n",buf);
			}
			if (add_cr) {
				register char	*s = buf, *d = cr_buf;
				register int	ll;
				for (ll = l; --ll >= 0;) {
					if ((*d++ = *s++) == '\n') {
						d[-1] = '\r';
						*d++ = '\n';
						++l;
					}
				}
				bufptr = cr_buf;
			} else {
				bufptr = buf;
			}
			retry = 0;
			if ((m = (stype == SOCK_STREAM) ? write(s,bufptr,l) : send(s,bufptr,l,0)) != l) {
				if (m < 0) goto barf;
				if (!quiet) fprintf(stderr,"server:  did not write entire buffer!\n");
				bufptr += m;
				l -= m;
				while (l && ++retry < MAXRETRY) {
					if (!quiet) fprintf(stderr,"server: retrying...\n");
					m = (stype == SOCK_STREAM) ? write(s,bufptr,l) : send(s,bufptr,l,0);
					if (m < 0) goto barf;
					bufptr += m;
					l -= m;
				}
				if (l) goto barf;
			}
			if (debug) fprintf(stderr,"sent\n");
		}
		break;
	}
	if (stype == SOCK_DGRAM || exit_on_eof) (void) kill(child,SIGKILL);
	return child;
barf:
	perror("server:write/send");
	(void) kill(child,SIGKILL);
	exit(1);
}

ckin(s,stype)
int s,stype;
{
	int	l, m, prev_was_cr = 0, retry;
	char	*bufptr;

#if	DEBUG
	fprintf(stderr,"Entered ckin, socket %x\n",s);
	if (stype == SOCK_STREAM) {
		fprintf(stderr,"SOCK_STREAM\n");
	} else {
		fprintf(stderr,"SOCK_DGRAM\n");
	}
#endif
#if	DEBUG
	fprintf(stderr,"STREAM/DGRAM: initial read\n");
#endif
	while ((l = (stype == SOCK_STREAM) ? read(s,buf,BUFSIZE) : recv(s,buf,BUFSIZ,0)) > 0) {
#if	DEBUG
		fprintf(stderr,"got LEN = %d\n",l);
#endif
		if (strip_cr) {
			register char	*s = buf, *d = cr_buf, ch;
			register int	ll;
			for (ll = l; --ll >= 0;) {
				if (prev_was_cr) {
					if ((ch = *d++ = *s++) != '\n') {
						*d++ = ch;
						d[-2] = '\r';
					}
					prev_was_cr = ch == '\r';
				} else {
					if ((*d++ = *s++) == '\r') {
						prev_was_cr = 1;
						--d;
					}
				}
			}
			l = d - cr_buf;
			bufptr = cr_buf;
		} else {
			bufptr = buf;
		}
		retry = 0;
		if ((m = write(1,bufptr,l)) != l) {
			if (m < 0) goto barf;
			bufptr += m;
			l -= m;
			while (l && ++retry < MAXRETRY) {
				if (!quiet) fprintf(stderr,"server: retrying...\n");
				m = write(1,bufptr,l);
				if (m < 0) goto barf;
				bufptr += m;
				l -= m;
			}
			if (l) goto barf;
		}
#if	DEBUG
		fprintf(stderr,"STREAM: reading");
#endif
	}
	if (strip_cr && prev_was_cr) write(1,"\r",1);
	if (debug) {
		fprintf(stderr,"server: ckin: EOF, errno %d\n",errno);
	}
	return;
barf:
	fprintf(stderr,"server:  writing to stdout failed!\n");
	exit(1);
}
